import datetime
import logging
import socket
from pathlib import Path
from typing import Iterable, Optional, Union

from pydantic import BaseModel
from typing_extensions import Literal

from cg.apps.cgstats.parsers.adapter_metrics import AdapterMetrics
from cg.apps.cgstats.parsers.conversion_stats import ConversionStats
from cg.apps.cgstats.parsers.quality_metrics import QualityMetrics
from cg.apps.cgstats.parsers.dragen_demultiplexing_stats import DragenDemultiplexingStats
from cg.apps.cgstats.parsers.run_info import RunInfo
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.constants.demultiplexing import DEMUX_STATS_PATH
from cg.models.demultiplex.flowcell import Flowcell

LOG = logging.getLogger(__name__)


class LogfileParameters(BaseModel):
    id_string: str  # This indicate software and version
    # This is the binary that was executed (atm only bcl2fastq)
    program: Literal["bcl2fastq", "dragen"] = "bcl2fastq"
    command_line: str
    time: datetime.datetime


class DemuxResults:
    """Class to gather information from a demultiplex result"""

    def __init__(self, demux_dir: Path, flowcell: Flowcell, bcl_converter: str):
        LOG.info("Instantiating DemuxResults with path %s", demux_dir)
        self.demux_dir: Path = demux_dir
        self.flowcell: Flowcell = flowcell
        self.bcl_converter = bcl_converter
        self._conversion_stats: Optional[ConversionStats] = None
        self._demultiplexing_stats: Optional[DragenDemultiplexingStats] = None
        self._adapter_metrics: Optional[AdapterMetrics] = None
        self._quality_metrics: Optional[QualityMetrics] = None
        self._runinfo: Optional[RunInfo] = None

    @property
    def run_name(self) -> str:
        return self.demux_dir.name

    @property
    def run_date(self) -> datetime.date:
        raw_time: str = self.run_name.split("_")[0]
        time_object: datetime.datetime = datetime.datetime.strptime(raw_time, "%y%m%d")
        return time_object.date()

    @property
    def machine_name(self) -> str:
        return self.run_name.split("_")[1]

    @property
    def demux_host(self) -> str:
        return socket.gethostname()

    @property
    def conversion_stats(self) -> ConversionStats:
        if self._conversion_stats:
            return self._conversion_stats
        self._conversion_stats = ConversionStats(self.conversion_stats_path)
        return self._conversion_stats

    @property
    def demultiplexing_stats(self) -> DragenDemultiplexingStats:
        if self._demultiplexing_stats:
            return self._demultiplexing_stats
        self._demultiplexing_stats = DragenDemultiplexingStats(self.demux_stats_path)
        return self._demultiplexing_stats

    @property
    def adapter_metrics(self) -> AdapterMetrics:
        if self._adapter_metrics:
            return self._adapter_metrics
        self._adapter_metrics = AdapterMetrics(self.adapter_metrics_path)
        return self._adapter_metrics

    @property
    def quality_metrics(self) -> QualityMetrics:
        if self._quality_metrics:
            return self._quality_metrics
        self._quality_metrics = QualityMetrics(self.quality_metrics_path)
        return self._quality_metrics

    @property
    def run_info(self) -> RunInfo:
        if self._runinfo:
            return self._runinfo
        self._runinfo = RunInfo(self.runinfo_path)
        return self._runinfo

    @property
    def conversion_stats_path(self) -> Union[Path, None]:
        return self.results_dir / DEMUX_STATS_PATH[self.bcl_converter]["conversion_stats"]

    @property
    def demux_stats_path(self) -> Path:
        return self.results_dir / DEMUX_STATS_PATH[self.bcl_converter]["demultiplexing_stats"]

    @property
    def adapter_metrics_path(self) -> Path:
        return self.results_dir / DEMUX_STATS_PATH[self.bcl_converter]["adapter_metrics_stats"]

    @property
    def quality_metrics_path(self) -> Path:
        return self.results_dir / DEMUX_STATS_PATH[self.bcl_converter]["quality_metrics"]

    @property
    def runinfo_path(self) -> Path:
        return self.results_dir / DEMUX_STATS_PATH[self.bcl_converter]["runinfo"]

    @property
    def stderr_log_path(self) -> Path:
        return DemultiplexingAPI.get_stderr_logfile(flowcell=self.flowcell)

    @property
    def stdout_log_path(self) -> Path:
        return DemultiplexingAPI.get_stdout_logfile(flowcell=self.flowcell)

    @property
    def results_dir(self) -> Path:
        return self.demux_dir / "Unaligned"

    @property
    def sample_sheet_path(self) -> Path:
        """Return the path to where the original sample sheet is"""
        return self.flowcell.sample_sheet_path

    @property
    def barcode_report(self) -> Path:
        """Return the path to the report with samples with low cluster count"""
        return self.demux_dir / "lane_barcode_summary.csv"

    @property
    def demux_sample_sheet_path(self) -> Path:
        """Return the path to sample sheet in demuxed flowcell dir"""
        return self.results_dir / self.flowcell.sample_sheet_path.name

    @property
    def copy_complete_path(self) -> Path:
        """Return the path to a file named copycomplete.txt used as flag that post processing is
        ready"""
        return self.demux_dir / "copycomplete.txt"

    @property
    def projects(self) -> Iterable[str]:
        """Return the project names created after demultiplexing

        How do we handle the Project_indexcheck directory?
        """
        for sub_dir in self.results_dir.iterdir():
            if not sub_dir.is_dir():
                continue
            if sub_dir.name.startswith("Project_"):
                project_name: str = sub_dir.name.split("_")[1]
                LOG.debug("Found project %s", project_name)
                yield project_name

    @property
    def raw_index_dir(self) -> Path:
        """Return the path to a index dir that is not given the 'Project_'-prefix"""
        return self.results_dir / "indexcheck"

    @property
    def raw_projects(self) -> Iterable[Path]:
        """Return the raw project names created after demultiplexing

        These are either 'indexcheck' or a number
        """
        for sub_dir in self.results_dir.iterdir():
            if not sub_dir.is_dir():
                LOG.debug("Skipping %s since it is not a directory", sub_dir)
                continue
            dir_name: str = sub_dir.name
            if dir_name in ["Stats", "Reports", "Logs"]:
                LOG.debug("Skipping %s dir %s", dir_name, sub_dir)
                continue
            if dir_name.startswith("Project_"):
                LOG.debug("Skipping already renamed project dir %s", sub_dir)
                continue
            if "index" in dir_name:
                LOG.debug("Skipping 'indexcheck' dir %s", sub_dir)
                continue
                # We now know that the rest of the directories are project directories
            yield sub_dir

    def files_renamed(self) -> bool:
        """Assert if the files have been renamed

        Check if the project files have been renamed, in that case it will not need post processing
        """
        return bool(list(self.projects))

    def get_logfile_parameters(self) -> LogfileParameters:
        get_logfile_parameters = {
            "bcl2fastq": self.get_bcl2fastq_logfile_parameters,
            "dragen": self.get_dragen_logfile_parameters,
        }

        return get_logfile_parameters[self.bcl_converter]()

    def get_dragen_logfile_parameters(self) -> LogfileParameters:
        err_log_path: Path = self.stderr_log_path
        out_log_path: Path = self.stdout_log_path

        LOG.info("Parse log file %s", err_log_path)

        if not err_log_path.exists():
            LOG.warning("Could not find log file %s", err_log_path)
            raise FileNotFoundError

        program: str = ""
        time: Optional[datetime.datetime] = None
        command_line: Optional[str] = None
        id_string: Optional[str] = None

        with open(err_log_path, "r") as logfile:
            for line in logfile.readlines():
                if "Dragen BCL Convert finished!" in line:
                    time: datetime.datetime = self._parse_time(line)

        LOG.info("Parse log file %s", out_log_path)

        if not out_log_path.exists():
            LOG.warning("Could not find log file %s", out_log_path)
            raise FileNotFoundError

        with open(out_log_path, "r") as logfile:
            for line in logfile.readlines():
                if "Command Line" in line:
                    line = line.strip()
                    split_line = line.split(" ")
                    program = split_line[2]
                    command_line = " ".join(split_line[2:])

                if "DRAGEN Host Software Version" in line:
                    line = line.strip()
                    split_line = line.split(" ")
                    id_string = split_line[4]

        return LogfileParameters(
            id_string=id_string, program=program, command_line=command_line, time=time
        )

    def get_bcl2fastq_logfile_parameters(self):

        log_path: Path = self.stderr_log_path

        LOG.info("Parse log file %s", log_path)

        if not log_path.exists():
            LOG.warning("Could not find log file %s", log_path)
            raise FileNotFoundError

        program: str = ""
        time: Optional[datetime.datetime] = None
        command_line: Optional[str] = None
        id_string: Optional[str] = None

        with open(log_path, "r") as logfile:
            for line in logfile.readlines():
                # Fetch the line where the call that was made is
                if "bcl2fastq" in line and "singularity" in line:
                    time: datetime.datetime = self._parse_time(line)

                    line = line.strip()
                    split_line = line.split(" ")
                    command_line: str = " ".join(split_line[1:])
                    program = split_line[6]  # get the executed program

                if "bcl2fastq v" in line:
                    id_string = line.strip()
                    break

        return LogfileParameters(
            id_string=id_string, program=program, command_line=command_line, time=time
        )

    @staticmethod
    def _parse_time(line: str) -> datetime.datetime:
        line.strip()
        split_line = line.split(" ")
        raw_time: str = split_line[0].strip("[]")
        try:
            time: datetime.datetime = datetime.datetime.strptime(raw_time, "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            time: datetime.datetime = datetime.datetime.strptime(raw_time, "%Y%m%d%H%M%S")

        return time

    def __str__(self):
        return f"DemuxResults(demux_dir={self.demux_dir},flowcell=Flowcell(flowcell_path={self.flowcell.path})"



""" Test adding files with Housekeeper API."""
from pathlib import Path
from typing import Dict, Any

from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.store_helpers import StoreHelpers


def test_add_file_with_flat_tag(
    hk_tag: str,
    housekeeper_api: MockHousekeeperAPI,
    helpers: StoreHelpers,
    hk_bundle_data: Dict[str, Any],
    fastq_file: Path,
):
    """Test that we can call hk with one existing tag."""

    # GIVEN an hk api populated with a version obj
    version_obj = helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)

    # GIVEN a tag that is just a string that exists
    assert housekeeper_api.tag(hk_tag)

    # WHEN we call add_file
    new_file = housekeeper_api.add_file(fastq_file, version_obj, hk_tag)

    # THEN the file should have been added to hk
    assert new_file


def test_add_file_with_list_of_tags(
    housekeeper_api: MockHousekeeperAPI,
    helpers: StoreHelpers,
    hk_bundle_data: Dict[str, Any],
    fastq_file: Path,
    not_existing_hk_tag: str,
):
    """Test that we can call Housekeeper with more than one tag."""

    # GIVEN an hk api populated with a version obj
    version_obj = helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)

    # GIVEN a list of tags that does not exist
    tags = [not_existing_hk_tag, not_existing_hk_tag + "_2"]
    for tag_name in tags:
        assert housekeeper_api.tag(tag_name) is None

    # WHEN we call add_file
    new_file = housekeeper_api.add_file(fastq_file, version_obj, tags)

    # THEN the file should have been added to Housekeeper
    assert new_file



"""Test for analysis"""
from typing import List

import pytest

from cg.constants import COLLABORATORS, Priority, GenePanelMasterList
from cg.constants.priority import SlurmQos
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.mip import MipAnalysisAPI


@pytest.mark.parametrize(
    "priority,expected_slurm_qos",
    [
        (Priority.clinical_trials, SlurmQos.NORMAL),
        (Priority.research, SlurmQos.LOW),
        (Priority.standard, SlurmQos.NORMAL),
        (Priority.priority, SlurmQos.HIGH),
        (Priority.express, SlurmQos.EXPRESS),
    ],
)
def test_get_slurm_qos_for_case(mocker, case_id: str, priority, expected_slurm_qos):
    """Test qet Quality of service (SLURM QOS) from the case priority"""

    # GIVEN a case that has a priority
    mocker.patch.object(AnalysisAPI, "get_priority_for_case")
    AnalysisAPI.get_priority_for_case.return_value = priority

    # When getting the SLURM QOS for the priority
    slurm_qos = AnalysisAPI.get_slurm_qos_for_case(AnalysisAPI, case_id=case_id)

    # THEN the expected slurm QOS should be returned
    assert slurm_qos is expected_slurm_qos


def test_gene_panels_correctly_added(customer_id):
    """Test get correct gene panel list."""

    # GIVEN a case that has a gene panel included in the gene panel master list
    default_panels_included: List[str] = [GenePanelMasterList.get_panel_names()[0]]
    master_list: List[str] = GenePanelMasterList.get_panel_names()

    # WHEN converting the gene panels between the default and the gene_panel_master_list
    list_of_gene_panels_used = MipAnalysisAPI.convert_panels(
        customer=customer_id, default_panels=default_panels_included
    )

    # THEN the list_of_gene_panels_used should return all gene panels
    assert set(list_of_gene_panels_used) == set(master_list)


def test_gene_panels_not_added(customer_id):
    """Test get only OMIM-AUTO and custom gene panel list."""
    # GIVEN a case that has a gene panel that is NOT included in the gene panel master list
    default_panels_not_included: List[str] = ["PANEL_NOT_IN_GENE_PANEL_MASTER_LIST"]

    # WHEN converting the gene panels between the default and the gene_panel_master_list
    list_of_gene_panels_used = MipAnalysisAPI.convert_panels(
        customer=customer_id, default_panels=default_panels_not_included
    )

    # THEN the list_of_gene_panels_used should return the custom panel and OMIM-AUTO
    assert set(list_of_gene_panels_used) == set(
        default_panels_not_included + [GenePanelMasterList.OMIM_AUTO]
    )



from typing import Optional, Any

from pydantic import BaseModel, constr, conlist

from cg.models.orders.constants import OrderType
from cg.models.orders.samples import (
    sample_class_for,
)
from cg.store import models


class OrderIn(BaseModel):
    name: constr(min_length=2, max_length=models.Sample.order.property.columns[0].type.length)
    comment: Optional[str]
    customer: constr(
        min_length=1, max_length=models.Customer.internal_id.property.columns[0].type.length
    )
    samples: conlist(Any, min_items=1)
    ticket: Optional[str]

    @classmethod
    def parse_obj(cls, obj: dict, project: OrderType):
        parsed_obj: OrderIn = super().parse_obj(obj)
        parsed_obj.parse_samples(project=project)
        return parsed_obj

    def parse_samples(self, project: OrderType):
        """
        Parses samples of by the type given by the project

        Parameters:
            project (OrderType): type of project

        Returns:
            Nothing
        """
        parsed_samples = []

        sample: dict
        for sample in self.samples:
            parsed_sample = sample_class_for(project=project).parse_obj(sample)
            parsed_samples.append(parsed_sample)

        self.samples = parsed_samples



import logging

from cg.apps.lims import LimsAPI
from cg.constants.scout_upload import BALSAMIC_CASE_TAGS, BALSAMIC_SAMPLE_TAGS
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.scout.scout_load_config import BalsamicLoadConfig, ScoutBalsamicIndividual
from cg.store import models
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


class BalsamicConfigBuilder(ScoutConfigBuilder):
    def __init__(
        self, hk_version_obj: hk_models.Version, analysis_obj: models.Analysis, lims_api: LimsAPI
    ):
        super().__init__(
            hk_version_obj=hk_version_obj, analysis_obj=analysis_obj, lims_api=lims_api
        )
        self.case_tags: CaseTags = CaseTags(**BALSAMIC_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**BALSAMIC_SAMPLE_TAGS)
        self.load_config: BalsamicLoadConfig = BalsamicLoadConfig(track="cancer")

    def include_case_files(self):
        LOG.info("Including BALSAMIC specific case level files")
        self.load_config.vcf_cancer = self.fetch_file_from_hk(self.case_tags.snv_vcf)
        self.load_config.vcf_cancer_sv = self.fetch_file_from_hk(self.case_tags.sv_vcf)
        self.include_cnv_report()
        self.include_multiqc_report()
        self.include_delivery_report()

    def include_sample_files(self, config_sample: ScoutBalsamicIndividual) -> None:
        LOG.info("Including BALSAMIC specific sample level files")

        sample_id: str = config_sample.sample_id
        if config_sample.alignment_path:
            if "tumor" in config_sample.alignment_path:
                sample_id = "tumor"
            elif "normal" in config_sample.alignment_path:
                sample_id = "normal"

        config_sample.vcf2cytosure = self.fetch_sample_file(
            hk_tags=self.sample_tags.vcf2cytosure, sample_id=sample_id
        )

    def build_config_sample(self, db_sample: models.FamilySample) -> ScoutBalsamicIndividual:
        """Build a sample with balsamic specific information"""
        config_sample = ScoutBalsamicIndividual()

        self.add_common_sample_info(config_sample=config_sample, db_sample=db_sample)
        if BalsamicAnalysisAPI.get_sample_type(db_sample.sample) == "tumor":
            config_sample.phenotype = "affected"
            config_sample.sample_id = "TUMOR"
        else:
            config_sample.phenotype = "unaffected"
            config_sample.sample_id = "NORMAL"

        config_sample.analysis_type = self.get_balsamic_analysis_type(db_sample.sample)

        return config_sample

    def get_balsamic_analysis_type(self, sample: models.Sample) -> str:
        """Returns a formatted balsamic analysis type"""

        analysis_type: str = BalsamicAnalysisAPI.get_application_type(sample_obj=sample)
        if analysis_type == "tgs":
            analysis_type = "panel"
        if analysis_type == "wgs":
            analysis_type = "wgs"

        return analysis_type

    def build_load_config(self) -> None:
        LOG.info("Build load config for balsamic case")
        self.add_common_info_to_load_config()
        self.load_config.human_genome_build = "37"
        self.load_config.rank_score_threshold = -100

        self.include_case_files()

        LOG.info("Building samples")
        db_sample: models.FamilySample

        for db_sample in self.analysis_obj.family.links:
            self.load_config.samples.append(self.build_config_sample(db_sample=db_sample))



import logging

from typing import List, Optional

from cg.apps.cgstats.crud.find import get_flowcell_id
from cg.apps.cgstats.db import models
from cg.apps.cgstats.stats import StatsAPI

log = logging.getLogger(__name__)


def delete_flowcell(manager: StatsAPI, flowcell_name: str):

    flowcell_id: Optional[int] = get_flowcell_id(flowcell_name=flowcell_name)

    if flowcell_id:
        flowcell: List[models.Flowcell] = manager.Flowcell.query.filter_by(
            flowcell_id=flowcell_id
        ).all()
        for entry in flowcell:
            log.info("Removing entry %s in from cgstats", entry.flowcellname)
            manager.delete(flowcell)
            manager.commit()



"""Tests for meta compress functionality that updates housekeeper"""

from pathlib import Path

from cg.constants import HK_FASTQ_TAGS
from cg.meta.compress import CompressAPI, files


def test_get_flow_cell_name(compress_api: CompressAPI, flowcell_name: str, flowcell_full_name: str):
    """Test functionality to extract the flow cell name from a run name given a designated structure"""

    # GIVEN a CompressAPI with a demux_root and a flowcell with a fastq in given demux_root
    fixture_flow_cell_name: str = flowcell_name
    fastq_path: Path = compress_api.demux_root.joinpath(
        Path(flowcell_full_name, "dummy_fastq.fastq.gz")
    )

    # WHEN retrieving the the name of the flow cell
    flow_cell_name: str = compress_api.get_flow_cell_name(fastq_path=fastq_path)

    # THEN the flow cell name retrieved should be identical to the fixture flow cell name used
    assert flow_cell_name == fixture_flow_cell_name


def test_add_fastq_housekeeper(
    compress_api,
    real_housekeeper_api,
    decompress_hk_spring_bundle,
    compression_files,
    store,
    helpers,
):
    """Test functionality to add fastq files to housekeeper

    This is done after a decompressed spring archives
    """
    # GIVEN real housekeeper api populated with a housekeeper bundle with spring info
    hk_bundle = decompress_hk_spring_bundle
    sample_id = hk_bundle["name"]
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    sample_obj = helpers.add_sample(store, internal_id=sample_id)
    compress_api.hk_api = real_housekeeper_api
    # GIVEN that there are no fastq files in HK
    version_obj = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    file_tags = set()
    for file_obj in version_obj.files:
        for tag in file_obj.tags:
            file_tags.add(tag.name)
    assert not set(HK_FASTQ_TAGS).intersection(file_tags)

    # WHEN adding the files to housekeeper
    compress_api.add_fastq_hk(
        sample_obj=sample_obj,
        fastq_first=compression_files.fastq_first_file,
        fastq_second=compression_files.fastq_second_file,
    )

    # THEN assert that the fastq files where added to HK
    version_obj = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    file_tags = set()
    for file_obj in version_obj.files:
        for tag in file_obj.tags:
            file_tags.add(tag.name)

    assert set(HK_FASTQ_TAGS).intersection(file_tags)


def test_add_decompressed_fastq(
    compress_api,
    real_housekeeper_api,
    decompress_hk_spring_bundle,
    compression_files,
    store,
    helpers,
):
    """Test functionality to add decompressed fastq files"""
    # GIVEN real housekeeper api populated with a housekeeper bundle with spring info
    hk_bundle = decompress_hk_spring_bundle
    sample_id = hk_bundle["name"]
    sample_obj = helpers.add_sample(store, internal_id=sample_id)
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    compress_api.hk_api = real_housekeeper_api
    # GIVEN that there exists a spring archive, spring metadata and unpacked fastqs
    version_obj = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    fastq_first = compression_files.fastq_first_file
    fastq_second = compression_files.fastq_second_file
    spring_file = compression_files.spring_file
    spring_metadata_file = compression_files.updated_spring_metadata_file
    assert files.is_file_in_version(version_obj=version_obj, path=spring_file)
    assert files.is_file_in_version(version_obj=version_obj, path=spring_metadata_file)

    assert not files.is_file_in_version(version_obj=version_obj, path=fastq_first)

    # WHEN adding decompresed files
    compress_api.add_decompressed_fastq(sample_obj=sample_obj)

    # THEN assert that the files where added
    version_obj = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    assert files.is_file_in_version(version_obj=version_obj, path=spring_file)
    assert files.is_file_in_version(version_obj=version_obj, path=spring_metadata_file)

    assert files.is_file_in_version(version_obj=version_obj, path=fastq_first)
    assert files.is_file_in_version(version_obj=version_obj, path=fastq_second)



"""Tests for cleaning demultiplexed-runs using
cg.meta.clean.demultiplexed_flow_cells.DemultiplexedRunsFlowCell"""
from pathlib import Path
from typing import Optional
from unittest import mock

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.meta.clean.demultiplexed_flow_cells import DemultiplexedRunsFlowCell
from cg.store import Store
from cg.store.models import Flowcell


@pytest.mark.parametrize(
    "flow_cell_path, result",
    [
        ("correct_flow_cell_path", True),
        ("incorrect_flow_cell_path_too_long", False),
        ("incorrect_flow_cell_path_extension", False),
    ],
)
@mock.patch("cg.apps.tb.api.TrailblazerAPI")
@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_flow_cell_name(
    mock_statusdb: Store,
    mock_hk: HousekeeperAPI,
    mock_tb: TrailblazerAPI,
    flow_cell_path: Path,
    result: bool,
    request,
):
    # GIVEN a flow cell
    flow_cell_path = request.getfixturevalue(flow_cell_path)
    mock_hk.files.return_value.count.return_value = 1
    flow_cell_obj = DemultiplexedRunsFlowCell(
        flow_cell_path=flow_cell_path,
        status_db=mock_statusdb,
        housekeeper_api=mock_hk,
        trailblazer_api=mock_tb,
    )

    # WHEN checking the name of the flow cell

    # THEN the attribute is_correctly_named should be set correctly
    assert flow_cell_obj.is_correctly_named == result


@pytest.mark.parametrize(
    "flow_cell_path, statusdb_return_value, result",
    [
        ("correct_flow_cell_path", Flowcell(), True),
        ("nonexistent_flow_cell_path", None, False),
    ],
)
@mock.patch("cg.apps.tb.api.TrailblazerAPI")
@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_flow_cell_exists_in_statusdb_(
    mock_statusdb: Store,
    mock_hk: HousekeeperAPI,
    mock_tb: TrailblazerAPI,
    flow_cell_path: Path,
    statusdb_return_value: Optional[Flowcell],
    result: bool,
    request,
):
    # GIVEN a flow cell that exists in statusdb
    flow_cell_path = request.getfixturevalue(flow_cell_path)
    mock_statusdb.flowcell.return_value = statusdb_return_value
    mock_hk.files.return_value.count.return_value = 1
    flow_cell_obj = DemultiplexedRunsFlowCell(
        flow_cell_path=flow_cell_path,
        status_db=mock_statusdb,
        housekeeper_api=mock_hk,
        trailblazer_api=mock_tb,
    )

    # WHEN checking if the flow cell exists in statusdb

    # THEN the attribute is_correctly_named should be set correctly
    assert flow_cell_obj.exists_in_statusdb == result


@mock.patch("cg.apps.tb.api.TrailblazerAPI")
@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_check_fastq_files_exist_in_hk(
    mock_statusdb: Store, mock_hk: HousekeeperAPI, mock_tb: TrailblazerAPI, correct_flow_cell_path
):
    # GIVEN a flow cell that has fastq files in housekeeper
    flow_cell_obj = DemultiplexedRunsFlowCell(
        flow_cell_path=correct_flow_cell_path,
        status_db=mock_statusdb,
        housekeeper_api=mock_hk,
        trailblazer_api=mock_tb,
    )
    flow_cell_obj._hk_fastq_files = ["file.fastq.gz"]

    # WHEN checking for fastq files in Housekeeper

    # THEN the check should return True
    assert flow_cell_obj.fastq_files_exist_in_housekeeper


@mock.patch("cg.apps.tb.api.TrailblazerAPI")
@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_check_no_fastq_files_exist_in_hk(
    mock_statusdb: Store, mock_hk: HousekeeperAPI, mock_tb: TrailblazerAPI, correct_flow_cell_path
):
    # GIVEN a flow cell that has no fastq files in housekeeper
    flow_cell_obj = DemultiplexedRunsFlowCell(
        flow_cell_path=correct_flow_cell_path,
        status_db=mock_statusdb,
        housekeeper_api=mock_hk,
        trailblazer_api=mock_tb,
    )
    flow_cell_obj._hk_fastq_files = []

    # WHEN checking for fastq files in Housekeeper

    # THEN the check set the property to False
    assert not flow_cell_obj.fastq_files_exist_in_housekeeper



"""Test CG CLI upload observations module."""

import logging

import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.apps.loqus import LoqusdbAPI
from cg.cli.upload.observations import observations
from cg.cli.upload.observations.utils import (
    get_observations_case,
    get_observations_case_to_upload,
    get_observations_api,
    get_observations_case_to_delete,
)
from cg.constants import EXIT_SUCCESS
from cg.constants.sequencing import SequencingMethod
from cg.constants.subject import PhenotypeStatus
from cg.exc import CaseNotFoundError
from cg.meta.upload.observations.observations_api import UploadObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from tests.store_helpers import StoreHelpers

from tests.apps.loqus.conftest import (
    fixture_loqusdbapi,
    fixture_loqus_config,
    loqus_process,
    loqus_binary_path,
    loqus_config_path,
    fixture_loqusdb_case_output,
)


def test_observations(
    base_context: CGConfig, cli_runner: CliRunner, helpers: StoreHelpers, caplog: LogCaptureFixture
):
    """Test upload of observations."""

    caplog.set_level(logging.DEBUG)
    store: Store = base_context.status_db

    # GIVEN a case ready to be uploaded to Loqusdb
    case: models.Family = helpers.add_case(store)
    case.customer.loqus_upload = True
    sample: models.Sample = helpers.add_sample(store, application_type=SequencingMethod.WES)
    store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)

    # WHEN trying to do a dry run upload to Loqusdb
    result = cli_runner.invoke(observations, [case.internal_id, "--dry-run"], obj=base_context)

    # THEN the execution should have been successful and stop at a dry run step
    assert result.exit_code == EXIT_SUCCESS
    assert "Dry run" in caplog.text


def test_get_observations_case(base_context: CGConfig, helpers: StoreHelpers):
    """Test get observations supported case."""

    store: Store = base_context.status_db

    # GIVEN an observations valid case
    case: models.Family = helpers.add_case(store)

    # WHEN retrieving a case given a specific case ID
    extracted_case = get_observations_case(base_context, case.internal_id, upload=True)

    # THEN the extracted case should match the stored one
    assert extracted_case == case


def test_get_observations_case_invalid_id(base_context: CGConfig, caplog: LogCaptureFixture):
    """Test get observations case providing an incorrect ID."""

    caplog.set_level(logging.DEBUG)

    # WHEN retrieving a case given a specific case ID
    with pytest.raises(CaseNotFoundError):
        # THEN a CaseNotFoundError should be raised
        get_observations_case(base_context, "invalid_case_id", upload=True)

    assert "Invalid case ID. Retrieving available cases for Loqusdb actions." in caplog.text


def test_get_observations_case_to_upload(base_context: CGConfig, helpers: StoreHelpers):
    """Test get case ready to be uploaded to Loqusdb."""

    store: Store = base_context.status_db

    # GIVEN a case ready to be uploaded to Loqusdb
    case: models.Family = helpers.add_case(store)
    case.customer.loqus_upload = True

    # WHEN retrieving a case given a specific case ID
    extracted_case = get_observations_case_to_upload(base_context, case.internal_id)

    # THEN the extracted case should match the stored one
    assert extracted_case == case


def test_get_observations_case_to_delete(
    base_context: CGConfig,
    helpers: StoreHelpers,
    loqusdbapi: LoqusdbAPI,
    loqusdb_case_output: bytes,
):
    """Test get case ready to be deleted from Loqusdb."""

    store: Store = base_context.status_db

    # GIVEN an observations instance and a case ready to be deleted from Loqusdb
    loqusdbapi.process.stdout = loqusdb_case_output.decode("utf-8").rstrip()
    base_context.loqusdb_api_ = loqusdbapi
    case: models.Family = helpers.add_case(store)

    # WHEN retrieving a case to delete given a specific case ID
    extracted_case = get_observations_case_to_delete(base_context, case.internal_id)

    # THEN the extracted case should match the stored one
    assert case == extracted_case


def test_get_observations_api(base_context: CGConfig, helpers: StoreHelpers):
    """Test get observation API given a Loqusdb supported case."""

    store: Store = base_context.status_db

    # GIVEN a Loqusdb supported case
    case: models.Family = helpers.add_case(store)
    sample: models.Sample = helpers.add_sample(store, application_type=SequencingMethod.WES)
    store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)

    # WHEN retrieving the observation API
    observations_api: UploadObservationsAPI = get_observations_api(base_context, case)

    # THEN the API should be returned
    assert observations_api



import datetime as dt

from cg.meta.transfer.lims import IncludeOptions, SampleState


def has_same_received_at(lims, sample_obj):
    lims_received_date = lims.get_received_date(lims_id=sample_obj.internal_id)
    return sample_obj.received_at == lims_received_date


def test_transfer_samples_received_at_overwriteable(transfer_lims_api):

    # GIVEN a sample exists in statusdb and it has a received_at date but no delivered_at date,
    # there is a sample in lims with the same internal id and another received_at date
    lims_api = transfer_lims_api.lims
    sample_store = transfer_lims_api.status
    assert sample_store.samples_to_deliver().count() > 0
    sample = sample_store.samples_to_deliver().first()
    assert sample.received_at
    new_date = dt.datetime.today()
    lims_samples = [
        sample_store.add_sample(
            name=sample.name, sex=sample.sex, internal_id=sample.internal_id, received=new_date
        )
    ]
    lims_api.mock_set_samples(lims_samples)
    assert not has_same_received_at(lims_api, sample)

    # WHEN transfer_samples has been called
    transfer_lims_api.transfer_samples(SampleState.RECEIVED, IncludeOptions.NOTINVOICED.value)

    # THEN the samples should have the same received_at as in lims
    assert has_same_received_at(lims_api, sample)


def test_transfer_samples_all(transfer_lims_api):

    # GIVEN a sample exists in statusdb and it has a received_at date but no delivered_at date,
    # there is a sample in lims with the same internal id and another received_at date
    lims_api = transfer_lims_api.lims
    sample_store = transfer_lims_api.status
    assert sample_store.samples_to_deliver().count() > 0
    sample = sample_store.samples_to_deliver().first()
    assert sample.received_at
    new_date = dt.datetime.today()
    lims_samples = [
        sample_store.add_sample(
            name=sample.name, sex=sample.sex, internal_id=sample.internal_id, received=new_date
        )
    ]
    lims_api.mock_set_samples(lims_samples)
    assert not has_same_received_at(lims_api, sample)

    # WHEN transfer_samples has been called
    transfer_lims_api.transfer_samples(SampleState.RECEIVED, IncludeOptions.ALL.value)

    # THEN the samples should have the same received_at as in lims
    assert has_same_received_at(lims_api, sample)


def test_transfer_samples_include_unset_received_at(transfer_lims_api):

    sample_store = transfer_lims_api.status
    samples = sample_store.samples()
    assert samples.count() >= 2

    # GIVEN sample with unset received_at
    untransfered_sample = samples[0]
    untransfered_sample.received_at = None
    untransfered_sample.preped_at = None
    untransfered_sample.sequenced_at = None
    untransfered_sample.delivered_at = None

    # GIVEN sample with set received_at
    transfered_sample = samples[1]
    transfered_sample.received_at = dt.datetime.today()
    transfered_sample.preped_at = None
    transfered_sample.sequenced_at = None
    transfered_sample.delivered_at = None

    # both samples has received date in lims
    lims_samples = []
    untransfered_sample_received_at_date = dt.datetime.today()
    transfered_sample_received_at_date = dt.datetime.today()
    lims_sample = sample_store.add_sample(
        name=untransfered_sample.name,
        sex=untransfered_sample.sex,
        internal_id=untransfered_sample.internal_id,
        received=untransfered_sample_received_at_date,
    )
    lims_samples.append(lims_sample)
    lims_sample = sample_store.add_sample(
        name=transfered_sample.name,
        sex=transfered_sample.sex,
        internal_id=transfered_sample.internal_id,
        received=transfered_sample_received_at_date,
    )
    lims_samples.append(lims_sample)
    lims_api = transfer_lims_api.lims
    lims_api.mock_set_samples(lims_samples)

    # WHEN calling transfer lims with include unset received_at
    transfer_lims_api.transfer_samples(SampleState.RECEIVED, IncludeOptions.UNSET.value)

    # THEN the sample that was not set has been set and the other sample was not touched
    assert has_same_received_at(lims_api, untransfered_sample)
    assert not has_same_received_at(lims_api, transfered_sample)



"""Fixtures for meta/upload tests"""
from typing import Dict

from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.coverage.api import ChanjoAPI
from cg.constants import Pipeline
from cg.constants.sequencing import SequencingMethod
from cg.constants.tags import HkMipAnalysisTag
from cg.meta.upload.coverage import UploadCoverageApi
from cg.meta.upload.gt import UploadGenotypesAPI
from cg.meta.upload.observations.observations_api import UploadObservationsAPI
from cg.store import Store, models
from tests.mocks.hk_mock import MockHousekeeperAPI


class MockCoverage(ChanjoAPI):
    """Mock chanjo coverage api."""


class MockLoqusAPI:
    """Mock LoqusAPI class."""

    def __init__(self, analysis_type=SequencingMethod.WGS):
        self.analysis_type = analysis_type

    @staticmethod
    def load(*args, **kwargs):
        """Mock load method."""
        _ = args
        _ = kwargs
        return dict(variants=12)

    @staticmethod
    def get_case(*args, **kwargs):
        """Mock get_case method."""
        _ = args
        _ = kwargs
        return {"case_id": "case_id", "_id": "123"}

    @staticmethod
    def get_duplicate(*args, **kwargs):
        """Mock get_duplicate method."""
        _ = args
        _ = kwargs
        return {"case_id": "case_id"}


@pytest.fixture(name="upload_genotypes_hk_bundle")
def fixture_upload_genotypes_hk_bundle(
    case_id: str, timestamp, case_qc_metrics_deliverables: Path, bcf_file: Path
) -> dict:
    """Returns a dictionary in Housekeeper format with files used in upload Genotype process."""
    data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(case_qc_metrics_deliverables),
                "archive": False,
                "tags": HkMipAnalysisTag.QC_METRICS,
            },
            {"path": str(bcf_file), "archive": False, "tags": ["snv-gbcf", "genotype"]},
        ],
    }
    return data


@pytest.fixture(name="analysis_obj")
def fixture_analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers
) -> models.Analysis:
    """Return a analysis object with a trio."""
    case_obj = analysis_store_trio.family(case_id)
    helpers.add_analysis(store=analysis_store_trio, case=case_obj, started_at=timestamp)
    return analysis_store_trio.family(case_id).analyses[0]


@pytest.fixture(name="upload_genotypes_api")
def fixture_upload_genotypes_api(
    real_housekeeper_api, genotype_api, upload_genotypes_hk_bundle, helpers
) -> UploadGenotypesAPI:
    """Create a upload genotypes api."""
    helpers.ensure_hk_bundle(real_housekeeper_api, upload_genotypes_hk_bundle, include=True)
    _api = UploadGenotypesAPI(
        hk_api=real_housekeeper_api,
        gt_api=genotype_api,
    )

    return _api


@pytest.fixture(scope="function")
def upload_observations_api(analysis_store, populated_housekeeper_api):
    """Create mocked UploadObservationsAPI object."""

    loqus_mock = MockLoqusAPI()

    _api = UploadObservationsAPI(
        status_api=analysis_store,
        hk_api=populated_housekeeper_api,
        loqus_api=loqus_mock,
    )

    yield _api


@pytest.fixture(scope="function")
def upload_observations_api_wes(analysis_store, populated_housekeeper_api):
    """Create mocked UploadObservationsAPI object."""

    loqus_mock = MockLoqusAPI(analysis_type=SequencingMethod.WES)

    _api = UploadObservationsAPI(
        status_api=analysis_store,
        hk_api=populated_housekeeper_api,
        loqus_api=loqus_mock,
    )

    yield _api


@pytest.yield_fixture(scope="function")
def coverage_upload_api(
    chanjo_config: Dict[str, Dict[str, str]], populated_housekeeper_api: MockHousekeeperAPI
):
    """Return a upload coverage API."""
    return UploadCoverageApi(
        status_api=None, hk_api=populated_housekeeper_api, chanjo_api=MockCoverage(chanjo_config)
    )


@pytest.fixture(scope="function")
def analysis(analysis_store, case_id, timestamp):
    """Fixture to mock an analysis."""
    _analysis = analysis_store.add_analysis(pipeline=Pipeline.BALSAMIC, version="version")
    _analysis.family = analysis_store.family(case_id)
    _analysis.config_path = "dummy_path"
    _analysis.completed_at = timestamp
    yield _analysis


@pytest.fixture(name="genotype_analysis_sex")
def fixture_genotype_analysis_sex() -> dict:
    """Return predicted sex per sample_id."""
    return {"ADM1": "male", "ADM2": "male", "ADM3": "female"}



"""Module to handle resetting of data in StatusDB."""

from typing import List

from cg.store import models


class LinkHelper:
    """Class that helps handle StatusDB links."""

    @staticmethod
    def is_all_samples_non_tumour(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are non tumour."""
        return all(not link.sample.is_tumour for link in links)

    @staticmethod
    def get_analysis_type_for_each_link(links: List[models.FamilySample]) -> list:
        """Return analysis type for each sample given by link list."""
        return [link.sample.application_version.application.analysis_type for link in links]



from alchy import Query

from cg.constants.subject import PhenotypeStatus
from cg.store.status_sample_filters import (
    get_samples_with_loqusdb_id,
    get_samples_without_loqusdb_id,
)


def test_get_samples_with_loqusdb_id(helpers, store, sample_store, sample_id, loqusdb_id):
    """Test sample extraction with Loqusdb ID."""

    # GIVEN a sample observations that has been uploaded to Loqusdb
    case = helpers.add_case(store)
    sample = helpers.add_sample(store, loqusdb_id=loqusdb_id)
    sample_not_uploaded = helpers.add_sample(store, internal_id=sample_id)
    sample_store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    sample_store.relate_sample(
        family=case, sample=sample_not_uploaded, status=PhenotypeStatus.UNKNOWN
    )

    # GIVEN a sample query
    samples: Query = store.samples()

    # WHEN retrieving the Loqusdb uploaded samples
    uploaded_samples = get_samples_with_loqusdb_id(samples=samples)

    # THEN the obtained sample should match the expected one
    assert sample in uploaded_samples
    assert sample_not_uploaded not in uploaded_samples


def test_get_samples_without_loqusdb_id(helpers, store, sample_store, sample_id, loqusdb_id):
    """Test sample extraction without Loqusdb ID."""

    # GIVEN a sample observations that has not been uploaded to Loqusdb
    case = helpers.add_case(store)
    sample = helpers.add_sample(store)
    sample_uploaded = helpers.add_sample(store, internal_id=sample_id, loqusdb_id=loqusdb_id)
    sample_store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    sample_store.relate_sample(family=case, sample=sample_uploaded, status=PhenotypeStatus.UNKNOWN)

    # GIVEN a sample query
    samples: Query = store.samples()

    # WHEN retrieving the Loqusdb not uploaded samples
    not_uploaded_samples = get_samples_without_loqusdb_id(samples=samples)

    # THEN the obtained sample should match the expected one
    assert sample in not_uploaded_samples
    assert sample_uploaded not in not_uploaded_samples



from alchy import Query
from cg.store import models


def get_samples_with_loqusdb_id(samples: Query, **kwargs) -> Query:
    """Fetches samples with a loqusdb ID."""
    return samples.filter(models.Sample.loqusdb_id.isnot(None))


def get_samples_without_loqusdb_id(samples: Query, **kwargs) -> Query:
    """Fetches samples without a loqusdb ID."""
    return samples.filter(models.Sample.loqusdb_id.is_(None))


def apply_sample_filter(function: str, samples: Query) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""
    filter_map = {
        "samples_uploaded_to_loqusdb": get_samples_with_loqusdb_id,
        "samples_not_uploaded_to_loqusdb": get_samples_without_loqusdb_id,
    }
    return filter_map[function](samples=samples)



"""Fixtures for cli compress functions"""

import datetime as dt
from pathlib import Path

import pytest

from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.compression import CompressionHkTags
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from tests.store_helpers import StoreHelpers


class MockCompressAPI(CompressAPI):
    """Mock out necessary functions for running the compress CLI functions."""

    def __init__(self):
        """initialize mock"""
        super().__init__(hk_api=None, crunchy_api=None, demux_root="")
        self.ntasks = 12
        self.mem = 50
        self.fastq_compression_success = True
        self.spring_decompression_success = True
        self.dry_run = False

    def set_dry_run(self, dry_run: bool):
        """Update dry run"""
        self.dry_run = dry_run

    def compress_fastq(self, sample_id: str, dry_run: bool = False):
        """Return if compression was succesfull"""
        _ = sample_id, dry_run
        return self.fastq_compression_success

    def decompress_spring(self, sample_id: str, dry_run: bool = False):
        """Return if decompression was succesfull"""
        _ = sample_id, dry_run
        return self.spring_decompression_success


@pytest.fixture(name="compress_api")
def fixture_compress_api():
    """Return a compress context"""
    return MockCompressAPI()


@pytest.fixture(scope="function", name="real_crunchy_api")
def fixture_real_crunchy_api(crunchy_config_dict):
    """crunchy api fixture"""
    _api = CrunchyAPI(crunchy_config_dict)
    _api.set_dry_run(True)
    yield _api


@pytest.fixture(scope="function", name="real_compress_api")
def fixture_real_compress_api(
    demultiplex_runs: Path, housekeeper_api: HousekeeperAPI, real_crunchy_api: CrunchyAPI
) -> CompressAPI:
    """Return a compress context"""
    return CompressAPI(
        crunchy_api=real_crunchy_api, hk_api=housekeeper_api, demux_root=demultiplex_runs.as_posix()
    )


@pytest.fixture(scope="function", name="real_populated_compress_fastq_api")
def fixture_real_populated_compress_fastq_api(
    real_compress_api: CompressAPI, compress_hk_fastq_bundle: dict, helpers: StoreHelpers
) -> CompressAPI:
    """Populated compress api fixture"""
    helpers.ensure_hk_bundle(real_compress_api.hk_api, compress_hk_fastq_bundle)

    return real_compress_api


@pytest.fixture(name="samples")
def fixture_samples():
    """Return a list of sample ids"""
    return ["sample1", "sample2", "sample3"]


# Store fixtures


class CaseInfo:
    """Holds information for creating a case"""

    def __init__(self, **kwargs):
        self.case_id = kwargs["case_id"]
        self.family_name = kwargs["family_name"]
        self.timestamp = kwargs["timestamp"]
        self.later_timestamp = kwargs["later_timestamp"]
        self.application_tag = kwargs["application_tag"]


@pytest.fixture(name="compress_case_info")
def fixture_compress_case_info(
    case_id,
    family_name,
    timestamp,
    later_timestamp,
    wgs_application_tag,
):
    """Returns a object with information about a case"""
    return CaseInfo(
        case_id=case_id,
        family_name=family_name,
        timestamp=timestamp,
        later_timestamp=later_timestamp,
        application_tag=wgs_application_tag,
    )


@pytest.fixture(name="populated_compress_store")
def fixture_populated_compress_store(store, helpers, compress_case_info, analysis_family):
    """Return a store populated with a completed analysis"""
    # Make sure that there is a case where anaylsis is completer
    helpers.ensure_case_from_dict(
        store,
        case_info=analysis_family,
        app_tag=compress_case_info.application_tag,
        ordered_at=compress_case_info.timestamp,
        completed_at=compress_case_info.later_timestamp,
    )

    return store


@pytest.fixture(name="populated_compress_multiple_store")
def fixture_populated_compress_multiple_store(
    store,
    helpers,
    compress_case_info,
    analysis_family,
):
    """Return a store populated with multiple completed analysis"""
    case_id = compress_case_info.case_id
    family_name = compress_case_info.family_name
    for number in range(10):
        analysis_family["internal_id"] = "_".join([str(number), case_id])
        analysis_family["name"] = "_".join([str(number), family_name])
        for ind, sample in enumerate(analysis_family["samples"]):
            analysis_family["samples"][ind]["internal_id"] = "_".join(
                [str(number), sample["internal_id"]]
            )
            if "father" in analysis_family["samples"][ind]:
                analysis_family["samples"][ind]["father"] = "_".join(
                    [str(number), analysis_family["samples"][ind]["father"]]
                )
            if "mother" in analysis_family["samples"][ind]:
                analysis_family["samples"][ind]["mother"] = "_".join(
                    [str(number), analysis_family["samples"][ind]["mother"]]
                )

        helpers.ensure_case_from_dict(
            store,
            case_info=analysis_family,
            app_tag=compress_case_info.application_tag,
            ordered_at=compress_case_info.timestamp,
            completed_at=compress_case_info.later_timestamp,
            created_at=dt.datetime.now() - dt.timedelta(days=1000),
        )
    return store


# Context fixtures
@pytest.fixture(name="compress_context")
def fixture_base_compress_context(
    compress_api: CompressAPI, store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context"""
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = store
    return cg_config_object


@pytest.fixture(name="store_fastq_context")
def fixture_store_fastq_context(
    compress_api: CompressAPI, store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context"""
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = store
    return cg_config_object


@pytest.fixture(name="populated_multiple_compress_context")
def fixture_populated_multiple_compress_context(
    compress_api: CompressAPI, populated_compress_multiple_store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context populated with a completed analysis"""
    # Make sure that there is a case where anaylis is completer
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = populated_compress_multiple_store
    return cg_config_object


@pytest.fixture(name="populated_compress_context")
def fixture_populated_compress_context(
    compress_api: CompressAPI, populated_compress_store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context populated with a completed analysis"""
    # Make sure that there is a case where analysis is completed
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object


@pytest.fixture(name="real_populated_compress_context")
def fixture_real_populated_compress_context(
    real_populated_compress_fastq_api: CompressAPI,
    populated_compress_store: Store,
    cg_config_object: CGConfig,
) -> CGConfig:
    """Return a compress context populated with a completed analysis"""
    # Make sure that there is a case where analysis is completed
    cg_config_object.meta_apis["compress_api"] = real_populated_compress_fastq_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object


# Bundle fixtures


@pytest.fixture(scope="function", name="sample")
def fixture_sample() -> str:
    """Return the sample id for first sample"""
    return "sample_1"


@pytest.fixture(scope="function", name="new_dir")
def fixture_new_dir(project_dir: Path) -> Path:
    """Return the path to a subdirectory"""
    new_dir = project_dir / "new_dir/"
    new_dir.mkdir()
    return new_dir


@pytest.fixture(name="spring_bundle")
def fixture_spring_bundle(project_dir, timestamp, sample):
    """Return a bundle with some spring info"""
    spring_file = project_dir / "file.spring"
    spring_file.touch()
    spring_meta_file = project_dir / "file.json"
    spring_meta_file.touch()
    hk_bundle_data = {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(spring_file),
                "archive": False,
                "tags": [sample, CompressionHkTags.SPRING],
            },
            {
                "path": str(spring_meta_file),
                "archive": False,
                "tags": [sample, CompressionHkTags.SPRING_METADATA],
            },
        ],
    }

    return hk_bundle_data


@pytest.fixture(name="spring_bundle_symlink_problem")
def fixture_spring_bundle_symlink_problem(project_dir, new_dir, timestamp, sample):
    """Setup an environment that is similar to the case we want to solve"""
    # Create spring files and faulty paths to spring files
    spring_file = project_dir / "file.spring"
    spring_file.touch()
    wrong_spring_file = new_dir / "file.spring"
    spring_meta_file = project_dir / "file.json"
    spring_meta_file.touch()
    wrong_spring_meta_file = new_dir / "file.json"
    hk_bundle_data = {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(wrong_spring_file),
                "archive": False,
                "tags": [sample, CompressionHkTags.SPRING],
            },
            {
                "path": str(wrong_spring_meta_file),
                "archive": False,
                "tags": [sample, CompressionHkTags.SPRING_METADATA],
            },
        ],
    }

    return hk_bundle_data


@pytest.fixture(name="symlinked_fastqs")
def fixture_symlinked_fastqs(project_dir: Path, new_dir: Path) -> dict:
    """Setup an environment that is similar to the case we want to solve"""
    fastq_first = project_dir / "first.fastq.gz"
    fastq_second = project_dir / "second.fastq.gz"

    symlinked_first = new_dir / "first.fastq.gz"
    symlinked_first.symlink_to(fastq_first)
    symlinked_second = new_dir / "second.fastq.gz"
    symlinked_second.symlink_to(fastq_second)
    return {
        "fastq_first": fastq_first,
        "fastq_second": fastq_second,
        "symlinked_first": symlinked_first,
        "symlinked_second": symlinked_second,
    }



import logging
from typing import Optional

import click
from alchy import Query
from cgmodels.cg.constants import Pipeline

from cg.apps.loqus import LoqusdbAPI
from cg.cli.upload.observations.utils import get_observations_case_to_delete
from cg.cli.workflow.commands import OPTION_LOQUSDB_SUPPORTED_PIPELINES, ARGUMENT_CASE_ID
from cg.exc import CaseNotFoundError
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from cg.constants.constants import DRY_RUN, SKIP_CONFIRMATION

LOG = logging.getLogger(__name__)


@click.command()
@ARGUMENT_CASE_ID
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_obj
def observations(context: CGConfig, case_id: str, yes: bool, dry_run: bool):
    """Delete a case from Loqusdb and reset the Loqus ID in StatusDB."""

    status_db: Store = context.status_db
    case: models.Family = get_observations_case_to_delete(context, case_id)

    if dry_run:
        LOG.info(f"Dry run: this would delete all variants in Loqusdb for case: {case.internal_id}")
        return

    LOG.info(f"This will delete all variants in Loqusdb for case: {case.internal_id}")
    if yes or click.confirm("Do you want to continue?", abort=True):
        context.loqusdb_api.delete_case(case_id=case_id)
        status_db.reset_loqusdb_observation_ids(case_id)
        status_db.commit()
        LOG.info(f"Removed observations for case: {case.internal_id}")


@click.command("available-observations")
@OPTION_LOQUSDB_SUPPORTED_PIPELINES
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_context
def available_observations(
    context: click.Context, pipeline: Optional[Pipeline], yes: bool, dry_run: bool
):
    """Delete available observation from Loqusdb."""

    status_db: Store = context.obj.status_db
    uploaded_observations: Query = status_db.observations_uploaded(pipeline)

    LOG.info(
        f"This would delete observations for the following cases: {[case.internal_id for case in uploaded_observations]}"
    )
    if yes or click.confirm("Do you want to continue?", abort=True):
        for case in uploaded_observations:
            try:
                context.invoke(observations, case_id=case.internal_id, yes=yes, dry_run=dry_run)
            except CaseNotFoundError:
                continue



from pathlib import Path
from typing import Any

from cg.constants.constants import FileFormat
from cg.io.json import read_json, write_json, write_json_stream, read_json_stream
from cg.io.yaml import read_yaml, write_yaml, read_yaml_stream, write_yaml_stream


class ReadFile:
    """Reading file using different methods"""

    read_file = {
        FileFormat.JSON: read_json,
        FileFormat.YAML: read_yaml,
    }

    @classmethod
    def get_content_from_file(cls, file_format: str, file_path: Path) -> Any:
        """Read file using file format dispatch table"""
        return cls.read_file[file_format](file_path=file_path)


class ReadStream:
    """Reading stream using different methods"""

    read_stream = {
        FileFormat.JSON: read_json_stream,
        FileFormat.YAML: read_yaml_stream,
    }

    @classmethod
    def get_content_from_stream(cls, file_format: str, stream: Any) -> Any:
        """Read stream using file format dispatch table"""
        return cls.read_stream[file_format](stream=stream)


class WriteFile:
    """write file using different methods"""

    write_file = {
        FileFormat.JSON: write_json,
        FileFormat.YAML: write_yaml,
    }

    @classmethod
    def write_file_from_content(cls, content: Any, file_format: str, file_path: Path) -> None:
        """write file using file format dispatch table"""
        cls.write_file[file_format](content=content, file_path=file_path)


class WriteStream:
    """write stream using different methods"""

    write_stream = {
        FileFormat.JSON: write_json_stream,
        FileFormat.YAML: write_yaml_stream,
    }

    @classmethod
    def write_stream_from_content(cls, content: Any, file_format: str) -> str:
        """write stream using file format dispatch table"""
        return cls.write_stream[file_format](content=content)



import datetime as dt
from typing import List

from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery, GenePanelMasterList
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import StatusEnum
from cg.store import models
from cg.constants.priority import Priority


class FastqSubmitter(Submitter):
    def submit_order(self, order: OrderIn) -> dict:
        """Submit a batch of samples for FASTQ delivery."""

        project_data, lims_map = process_lims(
            lims_api=self.lims, lims_order=order, new_samples=order.samples
        )
        status_data = self.order_to_status(order)
        self._fill_in_sample_ids(status_data["samples"], lims_map)
        new_samples = self.store_items_in_status(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket=order.ticket,
            items=status_data["samples"],
        )
        self._add_missing_reads(new_samples)
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input to status for fastq-only orders."""
        status_data = {
            "customer": order.customer,
            "order": order.name,
            "samples": [
                {
                    "application": sample.application,
                    "capture_kit": sample.capture_kit,
                    "comment": sample.comment,
                    "data_analysis": sample.data_analysis,
                    "data_delivery": sample.data_delivery,
                    "name": sample.name,
                    "priority": sample.priority,
                    "sex": sample.sex,
                    "tumour": sample.tumour,
                    "volume": sample.volume,
                }
                for sample in order.samples
            ],
        }
        return status_data

    def create_maf_case(self, sample_obj: models.Sample):
        case_obj: models.Family = self.status.add_case(
            data_analysis=Pipeline(Pipeline.MIP_DNA),
            data_delivery=DataDelivery(DataDelivery.ANALYSIS_FILES),
            name="_".join([sample_obj.name, "MAF"]),
            panels=[GenePanelMasterList.OMIM_AUTO],
            priority=Priority.research,
            ticket=sample_obj.original_ticket,
        )
        case_obj.customer = self.status.customer("cust000")
        relationship: models.FamilySample = self.status.relate_sample(
            family=case_obj, sample=sample_obj, status=StatusEnum.unknown
        )
        self.status.add(case_obj, relationship)

    def store_items_in_status(
        self, customer: str, order: str, ordered: dt.datetime, ticket: str, items: List[dict]
    ) -> List[models.Sample]:
        """Store fastq samples in the status database including family connection and delivery"""
        customer_obj = self.status.customer(customer)
        if customer_obj is None:
            raise OrderError(f"unknown customer: {customer}")
        new_samples = []
        case_obj: models.Family = self.status.find_family(customer=customer_obj, name=ticket)
        case: dict = items[0]
        with self.status.session.no_autoflush:
            for sample in items:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    sex=sample["sex"] or "unknown",
                    comment=sample["comment"],
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket,
                    priority=sample["priority"],
                    tumour=sample["tumour"],
                    capture_kit=sample["capture_kit"],
                )
                new_sample.customer = customer_obj
                application_tag = sample["application"]
                application_version = self.status.current_application_version(application_tag)
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version = application_version
                new_samples.append(new_sample)
                if not case_obj:
                    case_obj = self.status.add_case(
                        data_analysis=Pipeline(case["data_analysis"]),
                        data_delivery=DataDelivery(case["data_delivery"]),
                        name=ticket,
                        panels=None,
                        priority=case["priority"],
                        ticket=ticket,
                    )
                if (
                    not new_sample.is_tumour
                    and new_sample.application_version.application.prep_category == "wgs"
                ):
                    self.create_maf_case(sample_obj=new_sample)
                case_obj.customer = customer_obj
                new_relationship = self.status.relate_sample(
                    family=case_obj, sample=new_sample, status=StatusEnum.unknown
                )
                new_delivery = self.status.add_delivery(destination="caesar", sample=new_sample)
                self.status.add(case_obj, new_relationship, new_delivery)

        self.status.add_commit(new_samples)
        return new_samples



import logging
from pathlib import Path

import click
from cg.apps.cgstats.parsers.conversion_stats import ConversionStats
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.demux_report import create_demux_report
from cg.exc import FlowcellError
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flowcell import Flowcell

LOG = logging.getLogger(__name__)


@click.command(name="report")
@click.argument("flowcell-name")
@click.pass_obj
def create_report_cmd(context: CGConfig, flowcell_name: str):
    """Generate a demux report and print to stdout"""
    LOG.info("Check demuxed flowcell %s", flowcell_name)
    demux_api: DemultiplexingAPI = context.demultiplex_api
    try:
        flowcell: Flowcell = Flowcell(flowcell_path=demux_api.run_dir / flowcell_name)
    except FlowcellError:
        raise click.Abort
    demux_results: DemuxResults = DemuxResults(
        demux_dir=demux_api.out_dir / flowcell_name, flowcell=flowcell
    )
    conversion_stats: Path = demux_results.conversion_stats_path
    if not conversion_stats.exists():
        LOG.warning("Could not find conversion stats file %s", conversion_stats)
        raise click.Abort
    report = create_demux_report(
        conversion_stats=ConversionStats(demux_results.conversion_stats_path)
    )
    click.echo("\n".join(report))



import logging
from typing import Dict, List

from alchy import Query

from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import LOQUSDB_SUPPORTED_PIPELINES
from cg.constants.sequencing import SequencingMethod
from cg.exc import CaseNotFoundError, CustomerPermissionError, DataIntegrityError
from cg.meta.upload.observations.observations_api import UploadObservationsAPI
from cg.store import models, Store

from cg.models.cg_config import CGConfig
from cg.store.api.link import LinkHelper

LOG = logging.getLogger(__name__)


def get_observations_case(context: CGConfig, case_id: str, upload: bool) -> models.Family:
    """Return a verified Loqusdb case."""

    status_db: Store = context.status_db
    case: models.Family = status_db.family(case_id)
    if not case or case.data_analysis not in LOQUSDB_SUPPORTED_PIPELINES:
        LOG.error("Invalid case ID. Retrieving available cases for Loqusdb actions.")
        cases_to_process: Query = (
            status_db.observations_to_upload() if upload else status_db.observations_uploaded()
        )
        if not cases_to_process:
            LOG.info("There are no valid cases to be processed by Loqusdb")
        else:
            LOG.info("Provide one of the following case IDs: ")
            for case in cases_to_process:
                LOG.info(f"{case.internal_id} ({case.data_analysis})")

        raise CaseNotFoundError
    return case


def get_observations_case_to_upload(context: CGConfig, case_id: str) -> models.Family:
    """Return a verified case ready to be uploaded to Loqusdb."""

    case: models.Family = get_observations_case(context, case_id, upload=True)
    if not case.customer.loqus_upload:
        LOG.error(
            f"Customer {case.customer.internal_id} is not whitelisted for upload to Loqusdb. Canceling upload for "
            f"case {case.internal_id}."
        )
        raise CustomerPermissionError

    if not LinkHelper.is_all_samples_non_tumour(case.links):
        LOG.error(f"Case {case.internal_id} has tumor samples. Cancelling its upload.")
        raise DataIntegrityError
    return case


def get_observations_case_to_delete(context: CGConfig, case_id: str) -> models.Family:
    """Return a verified case ready to be deleted from Loqusdb."""

    loqusdb_api: LoqusdbAPI = context.loqusdb_api
    case: models.Family = get_observations_case(context, case_id, upload=False)
    if not loqusdb_api.case_exists(case_id):
        LOG.error(
            f"Case {case.internal_id} could not be found in Loqusdb",
        )
        raise CaseNotFoundError
    return case


def get_observations_api(context: CGConfig, case: models.Family) -> UploadObservationsAPI:
    """Return an observations API given a specific case object."""

    loqus_apis: Dict[SequencingMethod, LoqusdbAPI] = {
        SequencingMethod.WGS: LoqusdbAPI(context.dict()),
        SequencingMethod.WES: LoqusdbAPI(context.dict(), analysis_type=SequencingMethod.WES),
    }

    analysis_types: List[SequencingMethod] = LinkHelper.get_analysis_type_for_each_link(case.links)
    if len(set(analysis_types)) != 1 or analysis_types[0] not in (
        SequencingMethod.WES,
        SequencingMethod.WGS,
    ):
        LOG.error(
            f"Case {case.internal_id} has an undetermined analysis type or mixed analyses. Cancelling its upload."
        )
        raise DataIntegrityError

    analysis_type: SequencingMethod = analysis_types[0]
    return UploadObservationsAPI(
        status_api=context.status_db,
        hk_api=context.housekeeper_api,
        loqus_api=loqus_apis[analysis_type],
    )



import logging

import alchy

from cg.store import models
from cg.store.api.delete import DeleteDataHandler
from cg.store.api.findbusinessdata import FindBusinessDataHandler
from cg.store.api.reset import ResetHandler

from .add import AddHandler
from .findbasicdata import FindBasicDataHandler
from .status import StatusHandler

LOG = logging.getLogger(__name__)


class CoreHandler(
    AddHandler,
    DeleteDataHandler,
    FindBasicDataHandler,
    FindBusinessDataHandler,
    ResetHandler,
    StatusHandler,
):
    """Aggregating class for the store api handlers"""

    pass


class Store(alchy.Manager, CoreHandler):
    def __init__(self, uri):
        self.uri = uri
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=models.Model)



"""Import all constants here for easy access"""

from cg.constants.compression import FASTQ_DELTA, FASTQ_FIRST_READ_SUFFIX, FASTQ_SECOND_READ_SUFFIX
from cg.constants.constants import (
    CAPTUREKIT_CANCER_OPTIONS,
    CAPTUREKIT_OPTIONS,
    CASE_ACTIONS,
    COLLABORATORS,
    COMBOS,
    CONTAINER_OPTIONS,
    DEFAULT_CAPTURE_KIT,
    FLOWCELL_Q30_THRESHOLD,
    FLOWCELL_STATUS,
    PREP_CATEGORIES,
    SEX_OPTIONS,
    STATUS_OPTIONS,
    DataDelivery,
    FileExtensions,
    FlowCellStatus,
    HousekeeperTags,
    Pipeline,
)
from cg.constants.gene_panel import GenePanelMasterList
from cg.constants.paths import TMP_DIR
from cg.constants.priority import Priority
from cg.constants.process import EXIT_FAIL, EXIT_SUCCESS, RETURN_SUCCESS
from cg.constants.report import *
from cg.constants.sample_sources import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.constants.symbols import NO_PARENT, SINGLE_QUOTE, SPACE
from cg.constants.tags import HK_FASTQ_TAGS



"""Tests for store API status module."""

from alchy import Query

from cg.constants import Pipeline, Priority
from cg.constants.subject import PhenotypeStatus
from cg.store import Store, models
from cg.store.models import Application, Family, Sample
from tests.store_helpers import StoreHelpers


def test_samples_to_receive_external(sample_store, helpers):
    """Test fetching external sample."""
    store = sample_store
    # GIVEN a store with a mixture of samples
    assert store.samples().count() > 1

    # WHEN finding external samples to receive
    external_query = store.samples_to_receive(external=True)

    # THEN assert that only the external sample is returned
    assert external_query.count() == 1

    first_sample = external_query.first()
    # THEN assert that the sample is external in database
    assert first_sample.application_version.application.is_external is True
    # THEN assert that the sample is does not have a received at stamp
    assert first_sample.received_at is None


def test_samples_to_receive_internal(sample_store):
    # GIVEN a store with samples in a mix of states
    assert sample_store.samples().count() > 1
    assert len([sample for sample in sample_store.samples() if sample.received_at]) > 1

    # WHEN finding which samples are in queue to receive
    assert sample_store.samples_to_receive().count() == 1
    first_sample = sample_store.samples_to_receive().first()
    assert first_sample.application_version.application.is_external is False
    assert first_sample.received_at is None


def test_samples_to_sequence(sample_store):
    # GIVEN a store with sample in a mix of states
    assert sample_store.samples().count() > 1
    assert len([sample for sample in sample_store.samples() if sample.sequenced_at]) >= 1

    # WHEN finding which samples are in queue to be sequenced
    sequence_samples = sample_store.samples_to_sequence()

    # THEN it should list the received and partly sequenced samples
    assert sequence_samples.count() == 2
    assert {sample.name for sample in sequence_samples} == set(
        ["sequenced-partly", "received-prepared"]
    )
    for sample in sequence_samples:
        assert sample.sequenced_at is None
        if sample.name == "sequenced-partly":
            assert sample.reads > 0


def test_case_in_uploaded_observations(helpers: StoreHelpers, sample_store: Store, loqusdb_id: str):
    """Test retrieval of uploaded observations."""

    # GIVEN a case with observations that has been uploaded to Loqusdb
    analysis: models.Analysis = helpers.add_analysis(store=sample_store, pipeline=Pipeline.MIP_DNA)
    analysis.family.customer.loqus_upload = True
    sample: models.Sample = helpers.add_sample(sample_store, loqusdb_id=loqusdb_id)
    sample_store.relate_sample(analysis.family, sample, PhenotypeStatus.UNKNOWN)
    assert analysis.family.analyses
    for link in analysis.family.links:
        assert link.sample.loqusdb_id is not None

    # WHEN getting observations to upload
    uploaded_observations: Query = sample_store.observations_uploaded()

    # THEN the case should be in the returned query
    assert analysis.family in uploaded_observations


def test_case_not_in_uploaded_observations(helpers: StoreHelpers, sample_store: Store):
    """Test retrieval of uploaded observations that have not been uploaded to Loqusdb."""

    # GIVEN a case with observations that has not been uploaded to loqusdb
    analysis: models.Analysis = helpers.add_analysis(store=sample_store, pipeline=Pipeline.MIP_DNA)
    analysis.family.customer.loqus_upload = True
    sample: models.Sample = helpers.add_sample(sample_store)
    sample_store.relate_sample(analysis.family, sample, PhenotypeStatus.UNKNOWN)
    assert analysis.family.analyses
    for link in analysis.family.links:
        assert link.sample.loqusdb_id is None

    # WHEN getting observations to upload
    uploaded_observations: Query = sample_store.observations_uploaded()

    # THEN the case should not be in the returned query
    assert analysis.family not in uploaded_observations


def test_case_in_observations_to_upload(helpers: StoreHelpers, sample_store: Store):
    """Test extraction of ready to be uploaded to Loqusdb cases."""

    # GIVEN a case with completed analysis and samples w/o loqusdb_id
    analysis: models.Analysis = helpers.add_analysis(store=sample_store, pipeline=Pipeline.MIP_DNA)
    analysis.family.customer.loqus_upload = True
    sample: models.Sample = helpers.add_sample(sample_store)
    sample_store.relate_sample(analysis.family, sample, PhenotypeStatus.UNKNOWN)
    assert analysis.family.analyses
    for link in analysis.family.links:
        assert link.sample.loqusdb_id is None

    # WHEN getting observations to upload
    observations_to_upload: Query = sample_store.observations_to_upload()

    # THEN the case should be in the returned query
    assert analysis.family in observations_to_upload


def test_case_not_in_observations_to_upload(
    helpers: StoreHelpers, sample_store: Store, loqusdb_id: str
):
    """Test case extraction that should not be uploaded to Loqusdb."""

    # GIVEN a case with completed analysis and samples w loqusdb_id
    analysis: models.Analysis = helpers.add_analysis(store=sample_store, pipeline=Pipeline.MIP_DNA)
    analysis.family.customer.loqus_upload = True
    sample: models.Sample = helpers.add_sample(sample_store, loqusdb_id=loqusdb_id)
    sample_store.relate_sample(analysis.family, sample, PhenotypeStatus.UNKNOWN)
    assert analysis.family.analyses
    for link in analysis.family.links:
        assert link.sample.loqusdb_id is not None

    # WHEN getting observations to upload
    observations_to_upload: Query = sample_store.observations_to_upload()

    # THEN the case should not be in the returned query
    assert analysis.family not in observations_to_upload


def test_analyses_to_upload_when_not_completed_at(helpers, sample_store):
    """Test analyses to upload with no completed_at date."""
    # GIVEN a store with an analysis without a completed_at date
    helpers.add_analysis(store=sample_store)

    # WHEN fetching all analyses that are ready for upload
    records = [analysis_obj for analysis_obj in sample_store.analyses_to_upload()]

    # THEN no analysis object should be returned since they did not have a completed_at entry
    assert len(records) == 0


def test_analyses_to_upload_when_no_pipeline(helpers, sample_store, timestamp):
    """Test analyses to upload with no pipeline specified."""
    # GIVEN a store with one analysis
    helpers.add_analysis(store=sample_store, completed_at=timestamp)

    # WHEN fetching all analysis that are ready for upload without specifying pipeline
    records = [analysis_obj for analysis_obj in sample_store.analyses_to_upload(pipeline=None)]

    # THEN one analysis object should be returned
    assert len(records) == 1


def test_analyses_to_upload_when_analysis_has_pipeline(helpers, sample_store, timestamp):
    """Test analyses to upload to when existing pipeline."""
    # GIVEN a store with an analysis that has been run with MIP
    helpers.add_analysis(store=sample_store, completed_at=timestamp, pipeline=Pipeline.MIP_DNA)

    # WHEN fetching all analyses that are ready for upload and analysed with MIP
    records = [analysis_obj for analysis_obj in sample_store.analyses_to_upload(pipeline=None)]

    # THEN one analysis object should be returned
    assert len(records) == 1


def test_analyses_to_upload_when_filtering_with_pipeline(helpers, sample_store, timestamp):
    """Test analyses to upload to when existing pipeline and using it in filtering."""
    # GIVEN a store with an analysis that is analysed with MIP
    pipeline = Pipeline.MIP_DNA
    helpers.add_analysis(store=sample_store, completed_at=timestamp, pipeline=pipeline)

    # WHEN fetching all pipelines that are analysed with MIP
    records = [analysis_obj for analysis_obj in sample_store.analyses_to_upload(pipeline=pipeline)]

    for analysis_obj in records:
        # THEN pipeline should be MIP in the analysis object
        assert analysis_obj.pipeline == str(pipeline)


def test_analyses_to_upload_with_pipeline_and_no_complete_at(helpers, sample_store, timestamp):
    """Test analyses to upload to when existing pipeline and using it in filtering."""
    # GIVEN a store with an analysis that is analysed with MIP but does not have a completed_at
    pipeline = Pipeline.MIP_DNA
    helpers.add_analysis(store=sample_store, completed_at=None, pipeline=pipeline)

    # WHEN fetching all analyses that are ready for upload and analysed by MIP
    records = [analysis_obj for analysis_obj in sample_store.analyses_to_upload(pipeline=pipeline)]

    # THEN no analysis object should be returned since they where not completed
    assert len(records) == 0


def test_analyses_to_upload_when_filtering_with_missing_pipeline(helpers, sample_store, timestamp):
    """Test analyses to upload to when missing pipeline and using it in filtering."""
    # GIVEN a store with an analysis that has been analysed with "missing_pipeline"
    helpers.add_analysis(store=sample_store, completed_at=timestamp, pipeline=Pipeline.MIP_DNA)

    # WHEN fetching all analyses that was analysed with MIP
    records = [
        analysis_obj for analysis_obj in sample_store.analyses_to_upload(pipeline=Pipeline.FASTQ)
    ]

    # THEN no analysis object should be returned since there where no MIP analyses
    assert len(records) == 0


def test_multiple_analyses(analysis_store, helpers, timestamp_now, timestamp_yesterday):
    """Tests that analyses that are not latest are not returned."""

    # GIVEN an analysis that is not delivery reported but there exists a newer analysis
    case = helpers.add_case(analysis_store)
    analysis_oldest = helpers.add_analysis(
        analysis_store,
        case=case,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        delivery_reported_at=None,
    )
    analysis_store.add_commit(analysis_oldest)
    analysis_newest = helpers.add_analysis(
        analysis_store,
        case=case,
        started_at=timestamp_now,
        uploaded_at=timestamp_now,
        delivery_reported_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    analysis_store.relate_sample(
        family=analysis_oldest.family, sample=sample, status=PhenotypeStatus.UNKNOWN
    )

    # WHEN calling the analyses_to_delivery_report
    analyses = analysis_store.latest_analyses().all()

    # THEN only the newest analysis should be returned
    assert analysis_newest in analyses
    assert analysis_oldest not in analyses


def test_set_case_action(analysis_store, case_id):
    """Tests if actions of cases are changed to analyze."""
    # Given a store with a case with action None
    action = analysis_store.Family.query.filter(Family.internal_id == case_id).first().action

    assert action == None

    # When setting the case to "analyze"
    analysis_store.set_case_action(case_id=case_id, action="analyze")
    new_action = analysis_store.Family.query.filter(Family.internal_id == case_id).first().action

    # Then the action should be set to analyze
    assert new_action == "analyze"


def test_sequencing_qc_priority_express_sample_with_one_half_of_the_reads(
    base_store: Store, helpers, timestamp_now
):
    """Test if priority express sample(s), having more than 50% of the application target reads, pass sample QC."""

    # GIVEN a database with a case which has an express sample with half the amount of reads
    sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)
    application: Application = sample.application_version.application
    application.target_reads = 40
    sample.reads = 20
    sample.priority = Priority.express

    # WHEN retrieving the sequencing qc property of a the express sample
    sequencing_qc_ok: bool = sample.sequencing_qc

    # THEN the qc property should be True
    assert sequencing_qc_ok


def test_sequencing_qc_priority_standard_sample_with_one_half_of_the_reads(
    base_store: Store, helpers, timestamp_now
):
    """Test if priority standard sample(s), having more than 50% of the application target reads, pass sample QC."""

    # GIVEN a database with a case which has an normal sample with half the amount of reads
    sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)
    application: Application = sample.application_version.application
    application.target_reads = 40
    sample.reads = 20
    sample.priority = Priority.standard

    # WHEN retrieving the sequencing qc property of a the normal sample
    sequencing_qc_ok: bool = sample.sequencing_qc

    # THEN the qc property should be False
    assert not sequencing_qc_ok



"""Fixtures for the scout api tests"""

from pathlib import Path

import pytest

from cg.constants.constants import FileFormat
from cg.constants.pedigree import Pedigree
from cg.constants.subject import RelationshipStatus, Gender, PhenotypeStatus
from cg.io.controller import ReadFile
from tests.mocks.process_mock import ProcessMock

from cg.apps.scout.scoutapi import ScoutAPI


class MockScoutApi(ScoutAPI):
    def __init__(self, config: dict):
        binary_path = "scout"
        config_path = "config_path"
        self.process = ProcessMock(binary=binary_path, config=config_path)


@pytest.fixture(name="sample_dict")
def fixture_sample_dict() -> dict:
    sample_dict = {
        "analysis_type": "wgs",
        "bam_path": Path("path", "to", "sample.bam").as_posix(),
        "mt_bam": Path("path", "to", "reduced_mt.bam").as_posix(),
        "capture_kit": None,
        Pedigree.FATHER: RelationshipStatus.HAS_NO_PARENT,
        Pedigree.MOTHER: RelationshipStatus.HAS_NO_PARENT,
        "sample_id": "sample_id",
        "sample_name": "sample_name",
        Pedigree.SEX: Gender.MALE,
        "tissue_type": "unknown",
        Pedigree.PHENOTYPE: PhenotypeStatus.AFFECTED,
    }
    return dict(sample_dict)


@pytest.fixture(name="omim_disease_nr")
def fixture_omim_disease_nr() -> int:
    return 607208


@pytest.fixture(name="scout_dir")
def fixture_scout_dir(apps_dir: Path) -> Path:
    """Return the path to the scout fixtures dir"""
    return Path(apps_dir, "scout")


@pytest.fixture(name="causatives_file")
def fixture_causatives_file(scout_dir: Path) -> Path:
    """Return the path to a file with causatives output"""
    return Path(scout_dir, "export_causatives.json")


@pytest.fixture(name="cases_file")
def fixture_cases_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output"""
    return Path(scout_dir, "case_export.json")


@pytest.fixture(name="none_case_file")
def fixture_none_case_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output where mandatory fields are None"""
    return Path(scout_dir, "none_case_export.json")


@pytest.fixture(name="other_sex_case_file")
def fixture_other_sex_case_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output where sex is 'other0'"""
    return Path(scout_dir, "other_sex_case.json")


@pytest.fixture(name="none_case_raw")
def fixture_none_case_raw(none_case_file: Path) -> dict:
    """Return a single case of a export causatives run with scout"""
    cases: list = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=none_case_file
    )
    return cases[0]


@pytest.fixture(name="other_sex_case_output")
def fixture_other_sex_case_output(other_sex_case_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(other_sex_case_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="causative_output")
def fixture_causative_output(causatives_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(causatives_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="export_cases_output")
def fixture_export_cases_output(cases_file: Path) -> str:
    """Return the content of a export cases output"""
    with open(cases_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="scout_api")
def fixture_scout_api() -> ScoutAPI:
    return MockScoutApi({})



"""Tests the findbusinessdata part of the cg.store.api"""
from datetime import datetime

import pytest

from cg.store import Store, models
from cgmodels.cg.constants import Pipeline
from cg.constants.indexes import ListIndexes
from tests.store_helpers import StoreHelpers


def test_find_analysis_via_date(sample_store: Store, helpers: StoreHelpers):
    # GIVEN a case with an analysis with a startdate in the database
    analysis = helpers.add_analysis(store=sample_store, started_at=datetime.now())
    assert analysis.started_at

    # WHEN getting analysis via case_id and start date
    db_analysis = sample_store.analysis(analysis.family, analysis.started_at)

    # THEN the analysis should have been retrieved
    assert db_analysis == analysis


def test_get_latest_flow_cell_on_case(
    re_sequenced_sample_store: Store, case_id: str, flowcell_name: str
):
    """Test function to fetch the latest sequenced flowcell on a case"""

    # GIVEN a store with two flow cells in it, one being the latest sequenced of the two
    latest_flow_cell_obj: models.Flowcell = re_sequenced_sample_store.Flowcell.query.filter(
        models.Flowcell.name == flowcell_name
    ).first()

    # WHEN fetching the latest flow cell on a case with a sample that has been sequenced on both flow cells
    latest_flow_cell_on_case: models.Flowcell = (
        re_sequenced_sample_store.get_latest_flow_cell_on_case(family_id=case_id)
    )

    # THEN the fetched flow cell should have the same name as the other
    assert latest_flow_cell_obj.name == latest_flow_cell_on_case.name


def test_get_customer_id_from_ticket(analysis_store, customer_id, ticket: str):
    """Tests if the function in fact returns the correct customer"""
    # Given a store with a ticket

    # Then the function should return the customer connected to the ticket
    assert analysis_store.get_customer_id_from_ticket(ticket) == customer_id


def test_get_latest_ticket_from_case(case_id: str, analysis_store_single_case, ticket: str):
    """Tests if the correct ticket is returned for the given case"""
    # GIVEN a populated store with a case

    # WHEN the function is called
    ticket_from_case: str = analysis_store_single_case.get_latest_ticket_from_case(case_id=case_id)

    # THEN the ticket should be correct
    assert ticket == ticket_from_case


def test_get_case_pool(case_id: str, rml_pool_store: Store):
    """Tests if the correct pool is returned"""
    # GIVEN a case connected to a pool
    case_name = rml_pool_store.family(internal_id=case_id).name
    # WHEN the function is called
    pool = rml_pool_store.get_case_pool(case_id=case_id)
    # THEN the correct pool should be returned and the pool name should be the last part of the
    # case name
    assert pool.name == case_name.split("-", 1)[-1]


def test_get_ready_made_library_expected_reads(case_id: str, rml_pool_store: Store):
    """Test if the correct number of expected reads is returned."""

    # GIVEN a case with a sample with an application version
    application_version: models.ApplicationVersion = (
        rml_pool_store.family(case_id).links[ListIndexes.FIRST.value].sample.application_version
    )

    # WHEN the expected reads is fetched from the case
    expected_reads: int = rml_pool_store.get_ready_made_library_expected_reads(case_id=case_id)

    # THEN the fetched reads should be equal to the expected reads of the application versions application
    assert application_version.application.expected_reads == expected_reads


def test_get_application_by_case(case_id: str, rml_pool_store: Store):
    """Test that the correct application is returned on a case."""
    # GIVEN a case with a sample with an application version
    application_version: models.ApplicationVersion = (
        rml_pool_store.family(case_id).links[ListIndexes.FIRST.value].sample.application_version
    )

    # WHEN the application is fetched from the case
    application: models.Application = rml_pool_store.get_application_by_case(case_id=case_id)

    # THEN the fetched application should be equal to the application version application
    assert application_version.application == application



import datetime as dt
from typing import List

from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import SexEnum
from cg.models.orders.samples import RmlSample
from cg.store import models


class PoolSubmitter(Submitter):
    def validate_order(self, order: OrderIn) -> None:
        super().validate_order(order=order)
        self._validate_case_names_are_available(
            customer_id=order.customer, samples=order.samples, ticket=order.ticket
        )

    def submit_order(self, order: OrderIn) -> dict:
        status_data = self.order_to_status(order)
        project_data, lims_map = process_lims(
            lims_api=self.lims, lims_order=order, new_samples=order.samples
        )
        samples = [sample for pool in status_data["pools"] for sample in pool["samples"]]
        self._fill_in_sample_ids(samples, lims_map, id_key="internal_id")
        new_records = self.store_items_in_status(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket=order.ticket,
            items=status_data["pools"],
        )
        return {"project": project_data, "records": new_records}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert input to pools"""

        status_data = {
            "customer": order.customer,
            "order": order.name,
            "comment": order.comment,
            "pools": [],
        }

        # group pools
        pools = {}

        for sample in order.samples:
            pool_name = sample.pool
            application = sample.application
            data_analysis = sample.data_analysis
            data_delivery = sample.data_delivery
            priority = sample.priority

            if pool_name not in pools:
                pools[pool_name] = {}
                pools[pool_name]["name"] = pool_name
                pools[pool_name]["applications"] = set()
                pools[pool_name]["priorities"] = set()
                pools[pool_name]["samples"] = []

            pools[pool_name]["samples"].append(sample)
            pools[pool_name]["applications"].add(application)
            pools[pool_name]["priorities"].add(priority)

        # each pool must only have same of some values
        for pool in pools.values():

            applications = pool["applications"]
            priorities = pool["priorities"]
            pool_name = pool["name"]
            if len(applications) > 1:
                raise OrderError(f"different applications in pool: {pool_name} - {applications}")
            if len(priorities) > 1:
                raise OrderError(f"different priorities in pool: {pool_name} - {priorities}")

        for pool in pools.values():

            pool_name = pool["name"]
            applications = pool["applications"]
            application = applications.pop()
            pool_samples = pool["samples"]
            priorities = pool["priorities"]
            priority = priorities.pop()

            status_data["pools"].append(
                {
                    "name": pool_name,
                    "application": application,
                    "data_analysis": data_analysis,
                    "data_delivery": data_delivery,
                    "priority": priority,
                    "samples": [
                        {
                            "comment": sample.comment,
                            "control": sample.control,
                            "name": sample.name,
                        }
                        for sample in pool_samples
                    ],
                }
            )
        return status_data

    def store_items_in_status(
        self, customer: str, order: str, ordered: dt.datetime, ticket: str, items: List[dict]
    ) -> List[models.Pool]:
        """Store pools in the status database"""
        customer_obj: models.Customer = self.status.customer(customer)
        new_pools: List[models.Pool] = []
        new_samples: List[models.Sample] = []
        for pool in items:
            with self.status.session.no_autoflush:
                application_version: models.ApplicationVersion = (
                    self.status.current_application_version(pool["application"])
                )
            priority: str = pool["priority"]
            case_name: str = self.create_case_name(ticket=ticket, pool_name=pool["name"])
            case_obj: models.Family = self.status.find_family(customer=customer_obj, name=case_name)
            if not case_obj:
                data_analysis: Pipeline = Pipeline(pool["data_analysis"])
                data_delivery: DataDelivery = DataDelivery(pool["data_delivery"])
                case_obj = self.status.add_case(
                    data_analysis=data_analysis,
                    data_delivery=data_delivery,
                    name=case_name,
                    panels=None,
                    priority=priority,
                    ticket=ticket,
                )
                case_obj.customer = customer_obj
                self.status.add_commit(case_obj)

            new_pool: models.Pool = self.status.add_pool(
                application_version=application_version,
                customer=customer_obj,
                name=pool["name"],
                order=order,
                ordered=ordered,
                ticket=ticket,
            )
            sex: SexEnum = SexEnum.unknown
            for sample in pool["samples"]:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    sex=sex,
                    comment=sample["comment"],
                    control=sample.get("control"),
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket,
                    priority=priority,
                    application_version=application_version,
                    customer=customer_obj,
                    no_invoice=True,
                )
                new_samples.append(new_sample)
                self.status.relate_sample(family=case_obj, sample=new_sample, status="unknown")
            new_delivery = self.status.add_delivery(destination="caesar", pool=new_pool)
            self.status.add(new_delivery)
            new_pools.append(new_pool)
        self.status.add_commit(new_pools)

        return new_pools

    def _validate_case_names_are_available(
        self, customer_id: str, samples: List[RmlSample], ticket: str
    ):
        """Validate that the names of all pools are unused for all samples"""
        customer_obj: models.Customer = self.status.customer(customer_id)

        sample: RmlSample
        for sample in samples:
            case_name: str = self.create_case_name(pool_name=sample.pool, ticket=ticket)

            if self.status.find_family(customer=customer_obj, name=case_name):
                raise OrderError(
                    f"Case name {case_name} already in use for customer {customer_obj.name}"
                )

    @staticmethod
    def create_case_name(ticket: str, pool_name: str) -> str:
        return f"{ticket}-{pool_name}"



"""Test obseravtion input files"""
from pathlib import Path

import pytest


@pytest.fixture(name="file_does_not_exist")
def fixture_file_does_not_exist() -> Path:
    """Return a file path that does not exist"""
    return Path("dir", "does_not_really_exist")


@pytest.fixture(name="hk_tag")
def fixture_hk_tag() -> str:
    """Return a Housekeeper tag"""
    return "vcf"



"""
    API for compressing files. Functionality to compress FASTQ, decompress SPRING and clean files
"""

import logging
from pathlib import Path
from typing import List

from alchy import Query
from housekeeper.store import models as housekeeper_models

from cg.apps.crunchy import CrunchyAPI
from cg.apps.crunchy.files import update_metadata_date
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.compression import CompressionHkTags
from cg.meta.backup.backup import SpringBackupAPI
from cg.meta.compress import files
from cg.models import CompressionData, FileData
from cg.store import models
from cg.store.queries import get_cases_to_compress
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


class CompressAPI:
    """API for compressing BAM and FASTQ files."""

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        demux_root: str,
        crunchy_api: CrunchyAPI,
        backup_api: SpringBackupAPI = None,
        dry_run: bool = False,
    ):

        self.hk_api: HousekeeperAPI = hk_api
        self.crunchy_api: CrunchyAPI = crunchy_api
        self.backup_api: SpringBackupAPI = backup_api
        self.demux_root: Path = Path(demux_root)
        self.dry_run: bool = dry_run
        self.get_cases_to_compress: Query = get_cases_to_compress

    def set_dry_run(self, dry_run: bool):
        """Update dry run."""
        self.dry_run = dry_run
        if self.crunchy_api.dry_run is False:
            self.crunchy_api.set_dry_run(dry_run)
        if self.backup_api:
            self.backup_api.dry_run = self.dry_run

    def get_flow_cell_name(self, fastq_path: Path) -> str:
        """
        Extract the flow cell name from a fastq path assuming fastq files are kept in their
        demultiplexed path and the following run_name convention:

            - <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>:
            - Ex: 220128_A00689_0460_BHVN2FDSX2
        """
        run_name: str = fastq_path.relative_to(self.demux_root).parts[0]
        return run_name.split("_")[-1][1:]

    # Compression methods
    def compress_fastq(self, sample_id: str) -> bool:
        """Compress the FASTQ files for a individual."""
        LOG.info("Check if FASTQ compression is possible for %s", sample_id)
        version_obj: hk_models.Version = self.hk_api.get_latest_bundle_version(
            bundle_name=sample_id
        )
        if not version_obj:
            return False

        sample_fastq_dict = files.get_fastq_files(sample_id=sample_id, version_obj=version_obj)
        if not sample_fastq_dict:
            return False

        all_ok = True
        for run_name in sample_fastq_dict:
            LOG.info("Check if compression possible for run %s", run_name)
            compression_object = sample_fastq_dict[run_name]["compression_data"]
            if FileData.is_empty(compression_object.fastq_first):
                LOG.warning(
                    "Fastq files are empty for %s: %s", sample_id, compression_object.fastq_first
                )
                self.delete_fastq_housekeeper(
                    hk_fastq_first=sample_fastq_dict[run_name]["hk_first"],
                    hk_fastq_second=sample_fastq_dict[run_name]["hk_second"],
                )
                all_ok = False
                continue

            if not self.crunchy_api.is_fastq_compression_possible(compression_object):
                LOG.warning("FASTQ to SPRING not possible for %s, run %s", sample_id, run_name)
                all_ok = False
                continue

            LOG.info(
                "Compressing %s and %s for sample %s into SPRING format",
                compression_object.fastq_first,
                compression_object.fastq_second,
                sample_id,
            )
            self.crunchy_api.fastq_to_spring(compression_object, sample_id=sample_id)

        return all_ok

    def decompress_spring(self, sample_id: str) -> bool:
        """Decompress SPRING archive for a sample.

        This function will make sure that everything is ready for decompression from SPRING archive
        to FASTQ files.

            - Housekeeper will be updated to include FASTQ files
            - Housekeeper will still have the SPRING and SPRING metadata file
            - The SPRING metadata file will be updated to include date for decompression
            - PDC archived SPRING files will be retrieved and decrypted before decompression
        """
        version_obj: hk_models.Version = self.hk_api.get_latest_bundle_version(
            bundle_name=sample_id
        )
        if not version_obj:
            return False

        compression_objs = files.get_spring_paths(version_obj)
        for compression_obj in compression_objs:
            if not self.crunchy_api.is_spring_decompression_possible(compression_obj):
                LOG.info("SPRING to FASTQ decompression not possible for %s", sample_id)
                if not self.backup_api.is_to_be_retrieved_and_decrypted(
                    compression_obj.spring_path
                ):
                    return False
                LOG.info("Until the SPRING file is retrieved from PDC and decrypted")
                self.backup_api.retrieve_and_decrypt_spring_file(Path(compression_obj.spring_path))

            LOG.info(
                "Decompressing %s to FASTQ format for sample %s ",
                compression_obj.spring_path,
                sample_id,
            )

            self.crunchy_api.spring_to_fastq(compression_obj, sample_id=sample_id)
            update_metadata_date(compression_obj.spring_metadata_path)

        return True

    def clean_fastq(self, sample_id: str) -> bool:
        """Check that FASTQ compression is completed for a case and clean.

        This means removing compressed FASTQ files and update housekeeper to point to the new SPRING
        file and its metadata file.
        """
        LOG.info("Clean FASTQ files for %s", sample_id)
        version_obj: hk_models.Version = self.hk_api.get_latest_bundle_version(
            bundle_name=sample_id
        )
        if not version_obj:
            return False

        sample_fastq_dict = files.get_fastq_files(sample_id=sample_id, version_obj=version_obj)
        if not sample_fastq_dict:
            return False

        all_cleaned = True
        for run_name in sample_fastq_dict:

            compression_object = sample_fastq_dict[run_name]["compression_data"]

            if not self.crunchy_api.is_fastq_compression_done(compression_object):
                LOG.info("FASTQ compression not done for sample %s, run %s", sample_id, run_name)
                all_cleaned = False
                continue

            LOG.info("FASTQ compression done for sample %s, run %s!!", sample_id, run_name)

            self.update_fastq_hk(
                sample_id=sample_id,
                compression_obj=compression_object,
                hk_fastq_first=sample_fastq_dict[run_name]["hk_first"],
                hk_fastq_second=sample_fastq_dict[run_name]["hk_second"],
            )

            self.remove_fastq(
                fastq_first=compression_object.fastq_first,
                fastq_second=compression_object.fastq_second,
            )

        if all_cleaned:
            LOG.info("All FASTQ files cleaned for %s!!", sample_id)

        return all_cleaned

    def add_decompressed_fastq(self, sample_obj: models.Sample) -> bool:
        """Adds unpacked FASTQ files to housekeeper."""
        LOG.info("Adds FASTQ to Housekeeper for %s", sample_obj.internal_id)
        version_obj: hk_models.Version = self.hk_api.get_latest_bundle_version(
            bundle_name=sample_obj.internal_id
        )
        if not version_obj:
            return False

        spring_paths: List[CompressionData] = files.get_spring_paths(version_obj)
        if not spring_paths:
            LOG.warning("Could not find any spring paths for %s", sample_obj.internal_id)
        for compression_object in spring_paths:
            if not self.crunchy_api.is_spring_decompression_done(compression_object):
                LOG.info("SPRING to FASTQ decompression not finished %s", sample_obj.internal_id)
                return False

            fastq_first = compression_object.fastq_first
            fastq_second = compression_object.fastq_second
            if files.is_file_in_version(
                version_obj=version_obj, path=fastq_first
            ) or files.is_file_in_version(version_obj=version_obj, path=fastq_second):
                LOG.warning("FASTQ files already exists in Housekeeper")
                continue

            LOG.info(
                "Adding decompressed FASTQ files to Housekeeper for sample %s ",
                sample_obj.internal_id,
            )

            self.add_fastq_hk(
                sample_obj=sample_obj, fastq_first=fastq_first, fastq_second=fastq_second
            )

        return True

    def delete_fastq_housekeeper(
        self, hk_fastq_first: housekeeper_models.File, hk_fastq_second: housekeeper_models.File
    ) -> None:
        """Delete fastq files from housekeeper."""
        LOG.info("Deleting FASTQ files from Housekeeper")
        hk_fastq_first.delete()
        hk_fastq_second.delete()
        self.hk_api.commit()
        LOG.debug("FASTQ files deleted from Housekeeper")

    # Methods to update housekeeper
    def update_fastq_hk(
        self,
        sample_id: str,
        compression_obj: CompressionData,
        hk_fastq_first: housekeeper_models.File,
        hk_fastq_second: housekeeper_models.File,
    ) -> None:
        """Update Housekeeper with compressed FASTQ files and SPRING metadata file."""
        version_obj = self.hk_api.last_version(sample_id)

        spring_tags = [sample_id, CompressionHkTags.SPRING]
        spring_metadata_tags = [sample_id, CompressionHkTags.SPRING_METADATA]
        LOG.info("Updating FASTQ files in Housekeeper update for %s:", sample_id)
        LOG.info(
            "%s, %s -> %s, with tags %s",
            compression_obj.fastq_first,
            compression_obj.fastq_second,
            compression_obj.spring_path,
            spring_tags,
        )
        LOG.info(
            "Adds %s, with tags %s", compression_obj.spring_metadata_path, spring_metadata_tags
        )
        if self.dry_run:
            return

        LOG.info("Updating files in Housekeeper...")
        if files.is_file_in_version(version_obj, compression_obj.spring_path):
            LOG.info("SPRING file is already in Housekeeper")
        else:
            LOG.info("Adding SPRING file to Housekeeper")
            self.hk_api.add_file(
                path=compression_obj.spring_path, version_obj=version_obj, tags=spring_tags
            )
            self.hk_api.commit()

        if files.is_file_in_version(version_obj, compression_obj.spring_metadata_path):
            LOG.info("Spring metadata file is already in Housekeeper")
        else:
            self.hk_api.add_file(
                path=compression_obj.spring_metadata_path,
                version_obj=version_obj,
                tags=spring_metadata_tags,
            )
            self.hk_api.commit()

        self.delete_fastq_housekeeper(hk_fastq_first, hk_fastq_second)

    def add_fastq_hk(
        self, sample_obj: models.Sample, fastq_first: Path, fastq_second: Path
    ) -> None:
        """Add FASTQ files to housekeeper."""

        if not sample_obj.application_version.application.is_external:
            flow_cell_name: str = self.get_flow_cell_name(fastq_path=fastq_first)
            fastq_tags = [flow_cell_name, CompressionHkTags.FASTQ]
        else:
            fastq_tags = [sample_obj.internal_id, CompressionHkTags.FASTQ]
        last_version = self.hk_api.last_version(bundle=sample_obj.internal_id)
        LOG.info(
            "Adds %s, %s to bundle %s with tags %s",
            fastq_first,
            fastq_second,
            sample_obj.internal_id,
            fastq_tags,
        )
        if self.dry_run:
            return

        LOG.info("Updating files in Housekeeper...")
        self.hk_api.add_file(path=fastq_first, version_obj=last_version, tags=fastq_tags)
        self.hk_api.add_file(path=fastq_second, version_obj=last_version, tags=fastq_tags)
        self.hk_api.commit()

    # Methods to remove files from disc
    def remove_fastq(self, fastq_first: Path, fastq_second: Path):
        """Remove FASTQ files."""
        LOG.info("Will remove %s and %s", fastq_first, fastq_second)
        if self.dry_run:
            return
        fastq_first.unlink()
        LOG.debug("First FASTQ in pair %s removed", fastq_first)
        fastq_second.unlink()
        LOG.debug("Second FASTQ in pair %s removed", fastq_second)
        LOG.info("FASTQ files removed")
